@@ -7,27 +7,31 @@ from copy import copy
 from datetime import date, datetime
 from functools import wraps
 from itertools import chain, product
+from typing import Tuple, List, Dict, Set, Union, Any, Callable
+
+from six import with_metaclass
 
 from olo._speedups import decrypt_attrs, parse_attrs
 from olo.cache import CacheWrapper, delete_cache
 from olo.cached_query import CachedQuery
 from olo.compat import (str_types, iteritems, iterkeys, itervalues, izip,
-                        long, reduce, get_values, with_metaclass, xrange)
+                        long, reduce, get_values, xrange)
 from olo.context import Context, context, model_instantiate_context
-from olo.errors import DeparseError, ExpressionError, InvalidFieldError
+from olo.errors import DeparseError, ExpressionError, InvalidFieldError, ORMError
 from olo.events import after_delete, after_insert, after_update, before_update
-from olo.expression import BinaryExpression, Expression
+from olo.expression import Expression
 from olo.ext.exported import IS_EXPORTED_PROPERTY
 from olo.ext.n import N
-from olo.field import BaseField, DbField, Field, UnionField
+from olo.field import BaseField, DbField, Field, UnionField, BatchField
 from olo.funcs import RAND
 from olo.key import StrKey
 from olo.query import Query
-from olo.utils import (cached_property, camel2underscore, deprecation,
+from olo.statement import Assignment
+from olo.utils import (camel2underscore, deprecation,
                        friendly_repr, missing, override,
                        readonly_cached_property, type_checker)
 
-VALID_TYPES = str_types + (int, float, long, date)
+VALID_TYPES = str_types + (bytes, int, float, long, date, datetime, bool)
 PATTERN_N_NAME = re.compile('s$')
 
 
@@ -206,54 +210,10 @@ def init_field(field, cls, attr_name):
     field.AES_KEY = getattr(cls, 'AES_KEY', '')
 
 
-def _create_n_property(method, name):
-    default = method.default
-
-    def f(self):
-        session = self._olo_qs
-        if session is None:
-            entities = [self]  # pragma: no cover
-        else:
-            entities = session.entities
-
-        res = method(entities)
-
-        entity_mapping = {
-            e._get_singleness_pk_value(): e
-            for e in entities
-        }
-
-        if isinstance(res, dict):
-
-            for pv, item in iteritems(entity_mapping):
-                if hasattr(item, '_olo_qs'):
-                    setattr(item, name, res.get(pv, default))
-
-            return res.get(self._get_singleness_pk_value(), default)
-
-        if isinstance(res, list):
-
-            for idx, item in enumerate(entities):
-                if hasattr(item, '_olo_qs'):
-                    try:
-                        v = res[idx]
-                    except IndexError:
-                        v = default
-
-                    setattr(item, name, v)
-
-            try:
-                return res[self._olo_qs_idx]
-            except IndexError:  # pragma: no cover
-                return default  # pragma: no cover
-
-        return default  # pragma: no cover
-
-    f.__name__ = name
-    return cached_property(f)
-
-
-def _collect_fields(cls, attrs):
+def _collect_fields(cls, attrs) -> Tuple[
+    Set[str], Set[str], Set[Union[BaseField, Any]], Dict[str, Any], Set[str], Dict[str, Callable], Set[str], list, Dict[
+        Any, str], Dict[str, BaseField], Dict[str, BaseField]]:
+    # pylint: disable=too-many-statements
     fields = set()
     db_fields = set()
     primary_key = set()
@@ -264,6 +224,7 @@ def _collect_fields(cls, attrs):
     field_objs = []
     field_name_map = {}
     encrypted_fields = {}
+    setter_fields = {}
 
     for k in dir(cls) if not cls.__abstract__ else []:
         v = getattr(cls, k)
@@ -278,10 +239,12 @@ def _collect_fields(cls, attrs):
                     field_objs.append(v)
                 elif isinstance(v, DbField):
                     db_fields.add(k)
-                if isinstance(v, BaseField) and v.encrypt:
+                if v.encrypt:
                     encrypted_fields[k] = v
                 if v.choices is not None:
                     choices_field_sets.add(k)
+                if v._setter is not None:
+                    setter_fields[k] = v
             if v.is_primary_key():
                 primary_key.add(v)
             if v.on_update is not None:
@@ -300,7 +263,9 @@ def _collect_fields(cls, attrs):
             _name = PATTERN_N_NAME.sub('', name)
             if name == _name or _name in cls.__dict__:
                 continue  # pragma: no cover
-            setattr(cls, _name, _create_n_property(v, _name))
+            f = BatchField(object, default=v.default, name=_name)
+            f.getter(v.func)
+            setattr(cls, _name, f)
 
     ordered_field_attr_names = [
         f.attr_name for f in sorted(field_objs, key=lambda f: f.id)
@@ -310,7 +275,7 @@ def _collect_fields(cls, attrs):
         fields, db_fields, primary_key,
         on_updates, choices_field_sets, validates,
         exported_property_names, ordered_field_attr_names,
-        field_name_map, encrypted_fields,
+        field_name_map, encrypted_fields, setter_fields,
     )
 
 
@@ -372,6 +337,7 @@ class ModelMeta(type):
         return super(ModelMeta, mcs).__new__(mcs, class_name, bases, attrs)
 
     def __init__(cls, class_name, bases, attrs):
+        # pylint: disable=too-many-statements
         super(ModelMeta, cls).__init__(class_name, bases, attrs)
 
         if cls.__abstract__:
@@ -388,6 +354,7 @@ class ModelMeta(type):
             ordered_field_attr_names,
             field_name_map,
             encrypted_fields,
+            setter_fields,
         ) = _collect_fields(cls, attrs)
 
         cls.__fields__ = fields
@@ -399,6 +366,7 @@ class ModelMeta(type):
         cls.__ordered_field_attr_names__ = ordered_field_attr_names
         cls.__field_name_map__ = field_name_map
         cls.__encrypted_fields__ = encrypted_fields
+        cls.__setter_fields__ = setter_fields
 
         if '__table_name__' not in attrs:
             cls.__table_name__ = camel2underscore(class_name)
@@ -490,7 +458,7 @@ class ModelMeta(type):
         return cls._options.query_class(cls)
 
     @readonly_cached_property
-    def cache(cls):
+    def cache(cls) -> CacheWrapper:
         return cls._options.cache_class(cls)
 
     @readonly_cached_property
@@ -546,6 +514,13 @@ class Model(with_metaclass(ModelMeta)):
         self._init()
         self._data = attrs
 
+        if self._olo_is_new:
+            for k in self.__setter_fields__:
+                v = attrs.get(k, missing)
+                if v is missing:
+                    continue
+                setattr(self, k, v)
+
     def _init(self):
         self._parsed_data = {}
         self._dirty_fields = set()
@@ -650,6 +625,7 @@ class Model(with_metaclass(ModelMeta)):
         return is_success
 
     def update(self, **attrs):
+        # pylint: disable=too-many-statements
         self._check_attrs(attrs)
 
         attrs = self._wash_attrs(attrs)
@@ -664,6 +640,14 @@ class Model(with_metaclass(ModelMeta)):
             self._rollback()
             return False
 
+        for k in self.__setter_fields__:
+            v = attrs.get(k, missing)
+            if v is missing:
+                continue
+            f = getattr(self.__class__, k)
+            v = f._setter(self, v)
+            attrs[k] = v
+
         db = self._get_db()
 
         need_updates = {}
@@ -679,7 +663,7 @@ class Model(with_metaclass(ModelMeta)):
             need_updates[k] = res
 
         attrs = dict(need_updates, **attrs)
-        expressions, sql_attrs, db_attrs = self._split_attrs(attrs)
+        assignments, sql_attrs, db_attrs = self._split_attrs(attrs)
 
         sql_attrs = self._validate_attrs(sql_attrs, decrypt=False)
         db_attrs = self._validate_attrs(db_attrs, decrypt=False)
@@ -699,7 +683,7 @@ class Model(with_metaclass(ModelMeta)):
             self._rollback()
             return False
 
-        if expressions:
+        if assignments:
             expression = self.unique_expression
             if expression is None:
                 raise ExpressionError('Cannot update this instance because of '  # noqa pragma: no cover
@@ -710,21 +694,15 @@ class Model(with_metaclass(ModelMeta)):
                 'UPDATE',
                 ['TABLE', self._get_table_name()],
                 ['SET',
-                 ['SERIES'] + [exp.get_sql_ast() for exp in expressions]],
+                 ['SERIES'] + [asg.get_sql_ast() for asg in assignments]],
                 ['WHERE'] + [expression.get_sql_ast()]
             ]
 
-            try:
+            with db.transaction():
                 db.ast_execute(sql_ast)
-                if db.autocommit:
-                    db.commit()
-            except Exception:
-                if db.autocommit:
-                    db.rollback()  # pragma: no cover
-                raise
 
             dynamic_exps = [
-                exp for exp in expressions if isinstance(exp.right, Expression)
+                asg for asg in assignments if isinstance(asg.right, Expression)
             ]
             if dynamic_exps:
                 keys = list(map(lambda x: x.left.attr_name, dynamic_exps))
@@ -777,8 +755,10 @@ class Model(with_metaclass(ModelMeta)):
         return True
 
     def delete(self, **kwargs):
+        # before_delete will return None, so explicit compare with False
         if self.before_delete(**kwargs) is False:
-            return
+            return False
+
         expression = self.unique_expression
         if expression is None:
             raise ExpressionError('Cannot delete this instance because of '  # noqa pragma: no cover
@@ -792,16 +772,13 @@ class Model(with_metaclass(ModelMeta)):
         ]
 
         db = self._get_db()
-        db.ast_execute(sql_ast)
 
         def func():
             after_delete.send(self)
             self.after_delete(**kwargs)
 
-        if db.autocommit:
-            db.commit()
-            func()
-        else:
+        with db.transaction():
+            db.ast_execute(sql_ast)
             db.add_lazy_func(func)
 
         return True
@@ -850,8 +827,8 @@ class Model(with_metaclass(ModelMeta)):
         return getattr(cls, attr_name)
 
     @classmethod
-    def _split_attrs(cls, attrs, collect_expression=True):
-        expressions = []
+    def _split_attrs(cls, attrs, collect_assignment=True) -> Tuple[List[Assignment], Dict, Dict]:
+        assignments = []
         sql_attrs = {}
         db_attrs = {}
         for k, v in iteritems(attrs):
@@ -860,11 +837,11 @@ class Model(with_metaclass(ModelMeta)):
             elif k in cls.__fields__:
                 if not isinstance(v, Expression):
                     sql_attrs[k] = v
-                if collect_expression:
-                    f = getattr(cls, k)
+                if collect_assignment:
+                    f: Field = getattr(cls, k)
                     v = cls._deparse_attrs({k: v})[k]
-                    expressions.append(BinaryExpression(f, v, '='))
-        return expressions, sql_attrs, db_attrs
+                    assignments.append(Assignment(f, v))
+        return assignments, sql_attrs, db_attrs
 
     @classmethod
     def _check_attrs(cls, attrs):
@@ -914,6 +891,7 @@ class Model(with_metaclass(ModelMeta)):
 
         before_create_is_instance_method = getattr(self.before_create, '__self__', None) is self  # noqa pylint: disable=C
 
+        bcr = True
         if before_create_is_instance_method:
             bcr = self.before_create()
 
@@ -923,7 +901,8 @@ class Model(with_metaclass(ModelMeta)):
         if not before_create_is_instance_method:
             bcr = self.before_create(**attrs)  # pragma: no cover
 
-        if bcr is False:
+        # bcr will be none so must compare with False!!!
+        if bcr is False:  # noqa
             return False
 
         self._validate_attrs(attrs, parse=True,
@@ -931,33 +910,28 @@ class Model(with_metaclass(ModelMeta)):
 
         db = self._get_db()
 
-        expressions, _, _ = self._split_attrs(sql_attrs)
+        assignments, _, _ = self._split_attrs(sql_attrs)
 
-        if expressions:
+        if assignments:
             fields_ast = ['BRACKET']
             values_ast = ['VALUES']
 
-            for exp in expressions:
-                fields_ast.append(exp.left.get_sql_ast())
-                values_ast.append(['VALUE', exp.right])
+            for asg in assignments:
+                fields_ast.append(['QUOTE', asg.left.name])
+                values_ast.append(['VALUE', asg.right])
+
+            pk_name = self.get_singleness_pk_name()
 
             sql_ast = [
                 'INSERT',
                 ['TABLE', self._get_table_name()],
                 fields_ast,
-                values_ast
+                values_ast,
+                ['RETURNING', pk_name],
             ]
 
-            try:
+            with db.transaction():
                 id_ = db.ast_execute(sql_ast)
-                if db.autocommit:
-                    db.commit()
-            except Exception:
-                if db.autocommit:
-                    db.rollback()  # pragma: no cover
-                raise
-
-            pk_name = self.get_singleness_pk_name()
 
             if (
                     hasattr(self.__class__, pk_name) and
@@ -1039,7 +1013,7 @@ class Model(with_metaclass(ModelMeta)):
             return  # pragma: no cover
         pk_dict = self._get_pk_dict()
         if not pk_dict:
-            raise Exception('No pk dict!!!')  # pragma: no cover
+            raise ORMError('No pk dict!!!')  # pragma: no cover
         values = self.__class__.query(*missing_fields).filter(
             **pk_dict
         ).first()
@@ -1257,20 +1231,24 @@ class Model(with_metaclass(ModelMeta)):
         for k, v in iteritems(attrs):
             field = getattr(cls, k)
             is_field = isinstance(field, Field)
-            if not isinstance(v, Expression):
-                if v is not None or not field.noneable:
-                    if is_field and not isinstance(v, VALID_TYPES):
-                        v = field.deparse(v)
-                    if is_field and not isinstance(v, VALID_TYPES):
-                        raise DeparseError(  # pragma: no cover
-                            'The deparsed type of {}.{} is invalid. '
-                            'Type: {}; Value: {}. '
-                            'Please check the deparser of this field.'.format(
-                                cls.__name__, k, type(v), repr(v)
-                            )
-                        )
-                    v = field.encrypt_func(v) if field.encrypt else v
-                    v = field.input(v) if field.input else v
+            if isinstance(v, Expression):
+                res[k] = v
+                continue
+            if v is None and field.noneable:
+                res[k] = v
+                continue
+            if is_field and not isinstance(v, VALID_TYPES):
+                v = field.deparse(v)
+            if is_field and not isinstance(v, VALID_TYPES):
+                raise DeparseError(  # pragma: no cover
+                    'The deparsed type of {}.{} is invalid. '
+                    'Type: {}; Value: {}. '
+                    'Please check the deparser of this field.'.format(
+                        cls.__name__, k, type(v), repr(v)
+                    )
+                )
+            v = field.encrypt_func(v) if field.encrypt else v
+            v = field.input(v) if field.input else v
             res[k] = v
         return res
 
